from fhir.resources.bundle import BundleEntry
from typing import List
from service_identifier import ServiceIdentifier


class BundleContent:
    def __init__(self, service_identifier: ServiceIdentifier, focus_ressource_id: str, sender: str,  bundle_entries: List[BundleEntry]):
        self.service_identifier = service_identifier
        self.focus_ressource_id = focus_ressource_id
        self.sender = sender
        self.bundle_entries = bundle_entries



from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.messageheader import MessageHeaderDestination, MessageHeaderSource
from fhir.resources.fhirtypes import ReferenceType
from fhir.resources.operationoutcome import OperationOutcome
from typing import List
from uuid import uuid4

from app_transport_framework_library.ressource_creators.message_bundle_creator import MessageBundleCreator
from app_transport_framework_library.ressource_creators.message_header_creator import MessageHeaderCreator



class OperationOutcomeBundleCreator:
    def create_operation_outcome_receipt_bundle(message_sender: ReferenceType,
                                                source: MessageHeaderSource,
                                                destinations: List[MessageHeaderDestination],
                                                operation_outcome: OperationOutcome) -> Bundle:

        message_header = MessageHeaderCreator.create_message_header(
            str(uuid4()),
            message_sender,
            source,
            destinations,
            code_system="https://gematik.de/fhir/atf/CodeSystem/operation-identifier-cs",
            use_case="atf;Empfangsbestaetigung",
            use_case_display="Empfangsbestätigung und Auskunft über FHIR Interpretierbarkeit der Nachricht",
            focus_reference=operation_outcome.id)

        resources = [
            BundleEntry(
                fullUrl=f"urn:uuid:{operation_outcome.id}",
                resource=operation_outcome
            )
        ]

        bundle = MessageBundleCreator.create_message_bundle(
            message_header=message_header,
            resources=resources
        )

        return bundle



from typing import List, Tuple
from fhir.resources.bundle import Bundle
from fhir.resources.operationoutcome import OperationOutcome, OperationOutcomeIssue
from fhir.resources.messageheader import MessageHeader
from fhir.resources.bundle import BundleEntry

from app_transport_framework_library.base_use_case_validator import BaseUseCaseValidator
from app_transport_framework_library.models.empfangsbestaetigung import Empfangsbestaetigung


class EmpfangsbestaetigungHandler(BaseUseCaseValidator):

    def resolve_reference(self, reference_str: str, bundle: Bundle):
        for entry in bundle.entry:
            if entry.resource.id == reference_str.split('urn:uuid:')[-1]:
                return entry.resource
        return None

    def handle(self, message_header: MessageHeader, bundle: Bundle) -> Tuple[List[BundleEntry], List[OperationOutcomeIssue]]:
        self.bundleEntries = []
        self.issues = []
        message_header = self.get_ressource_by_type(bundle, MessageHeader)
        outcome = self.get_ressource_by_type(bundle, OperationOutcome)
        message_id = outcome.extension[0].valueString.split(
            'urn:uuid:')[-1]
        empfangsbestaetigung = Empfangsbestaetigung(
            message_id, True, message_header.sender.identifier.value, message_header.destination[0].receiver.identifier.value, outcome.issue[0])

        self.received_Empfangsbestaetigung_event.trigger(empfangsbestaetigung)
        return self.bundleEntries, self.issues

    def get_ressource_by_type(self, bundle: Bundle, type):
        return next((entry.resource for entry in bundle.entry if isinstance(
            entry.resource, type)), None)



from typing import List, Tuple
from fhir.resources.messageheader import MessageHeader
from fhir.resources.bundle import Bundle
from fhir.resources.bundle import Bundle
from fhir.resources.operationoutcome import OperationOutcome, OperationOutcomeIssue
from fhir.resources.messageheader import MessageHeader
from fhir.resources.communication import Communication
from fhir.resources.bundle import BundleEntry
from app_transport_framework_library.base_use_case_validator import BaseUseCaseValidator

class SelbsttestLieferungValidator(BaseUseCaseValidator):
    def resolve_reference(self, reference_str: str, bundle: Bundle):
        for entry in bundle.entry:
            if entry.resource.id == reference_str.split('urn:uuid:')[-1]:
                return entry.resource
        return None

    def handle(self, message_header: MessageHeader, bundle: Bundle) -> Tuple[List[BundleEntry], List[OperationOutcomeIssue]]:
        self.bundleEntries = []
        self.issues = []
        if not any([isinstance(self.resolve_reference(focus_ref.reference, Bundle.parse_raw(bundle.json())), Communication) for focus_ref in message_header.focus]):
            self.issues.append(
                OperationOutcomeIssue(
                    severity="fatal",
                    code="invalid",
                    diagnostics="Eine Communication-Ressource wurde im MessageHeader.focus nicht gefunden"
                )
            )
        else:
            self.issues.append(
                OperationOutcomeIssue(
                    severity="information",
                    code="informational",
                    diagnostics="Anfrage erfolgreich entgegengenommen"
                )
            )

            for entry in bundle.entry:
                if not isinstance(entry.resource, (MessageHeader, OperationOutcome)):
                    self.bundleEntries.append(entry.resource)

        return self.bundleEntries, self.issues



from fhir.resources.bundle import Bundle


class MessageToSend:
    def __init__(self, atf_bundle: Bundle, receiver: str, message_type: str):
        self.atf_bundle = atf_bundle
        self.receiver = receiver
        self.message_type = message_type



from fhir.resources.operationoutcome import OperationOutcome, OperationOutcomeIssue
from fhir.resources.meta import Meta
from fhir.resources.extension import Extension
from uuid import uuid4


class OperationOutcomeCreator:
    def create_operation_outcome_ressource(message_id: str, issues: list[OperationOutcomeIssue]) -> OperationOutcome:
        operation_outcome = OperationOutcome(
            id=str(uuid4()),
            meta=Meta.construct(
                profile=["https://gematik.de/fhir/atf/StructureDefinition/atf-operation-outcome"]),
            extension=[
                Extension(
                    url="https://gematik.de/fhir/atf/StructureDefinition/atf-message-id-ex",
                    valueString=message_id
                )
            ],
            issue=issues
        )
        return operation_outcome



class Event:
    def __init__(self):
        self._subscribers = []

    def subscribe(self, callback):
        self._subscribers.append(callback)

    def unsubscribe(self, callback):
        self._subscribers.remove(callback)

    def trigger(self, *args, **kwargs):
        for callback in self._subscribers:
            callback(*args, **kwargs)



from abc import ABC, abstractmethod
from typing import List, Tuple
from fhir.resources.messageheader import MessageHeader, MessageHeaderSource
from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.fhirtypes import ReferenceType
from fhir.resources.operationoutcome import OperationOutcomeIssue
from app_transport_framework_library.models.event import Event

class BaseUseCaseValidator(ABC):
    def __init__(self, sender: ReferenceType, source: MessageHeaderSource):
        self.sender = sender
        self.source = source
        self.bundleEntries = []
        self.issues = []
        self.received_Empfangsbestaetigung_event = Event()

    @abstractmethod
    def handle(self, message_header: MessageHeader, bundle: Bundle) -> Tuple[List[BundleEntry], List[OperationOutcomeIssue]]:
        pass



from datetime import datetime, timezone
from typing import List
from fhir.resources.meta import Meta
from fhir.resources.bundle import Bundle, BundleEntry
from fhir.resources.messageheader import MessageHeader
from fhir.resources.fhirtypes import IdentifierType, Instant
from uuid import uuid4


class MessageBundleCreator:
    @staticmethod
    def create_message_bundle(message_header: MessageHeader, resources: List[BundleEntry]) -> Bundle:
        bundle_id = str(uuid4())

        entries = [
            BundleEntry(
                fullUrl=f"urn:uuid:{message_header.id}", resource=message_header)
        ]

        entries.extend(resources)

        message_bundle = Bundle(
            id=bundle_id,
            meta=Meta.construct(profile=[
                "https://gematik.de/fhir/atf/StructureDefinition/bundle-app-transport-framework"
            ]),
            type="message",
            identifier=IdentifierType(
                system="urn:ietf:rfc:3986", value=f"urn:uuid:{bundle_id}"),
            entry=entries,
            timestamp=datetime.now(timezone.utc).isoformat(),
        )

        return message_bundle



import logging
from json import JSONDecodeError
from fhir.resources.bundle import Bundle
from fhir.resources.messageheader import MessageHeader, MessageHeaderSource, MessageHeaderDestination
from fhir.resources.operationoutcome import OperationOutcomeIssue
from fhir.resources.fhirtypes import ReferenceType
from app_transport_framework_library.base_use_case_validator import BaseUseCaseValidator
from app_transport_framework_library.models.bundle_content import BundleContent
from app_transport_framework_library.models.event import Event
from app_transport_framework_library.models.message_to_send import MessageToSend
from app_transport_framework_library.models.service_identifier import ServiceIdentifier
from app_transport_framework_library.ressource_creators.operation_outcome_bundle_creator import OperationOutcomeBundleCreator
from app_transport_framework_library.ressource_creators.operation_outcome_creator import OperationOutcomeCreator
from app_transport_framework_library.use_cases.empfangsbestaetigung_handler import EmpfangsbestaetigungHandler
from app_transport_framework_library.use_cases.selbsttest_lieferung_validator import SelbsttestLieferungValidator

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class ATF_BundleProcessor:
    def __init__(self, sender: ReferenceType, source: MessageHeaderSource):
        self.sender = sender
        self.source = source
        self.use_case_handlers = {}
        self.message_to_send_event = Event()
        self.received_Empfangsbestaetigung_event = Event()
        self.bundle_content_to_process_event = Event()
        self.empfangsbestaetigungs_handler = EmpfangsbestaetigungHandler(
            self.sender, self.source)
        self.register_standard_use_case_handlers()
        self.empfangsbestaetigungs_handler.received_Empfangsbestaetigung_event.subscribe(
            self.on_empfangsbestaetigung_to_forwarded)

    def on_empfangsbestaetigung_to_forwarded(self, empfangsbestaetigung):
        self.received_Empfangsbestaetigung_event.trigger(empfangsbestaetigung)

    def register_standard_use_case_handlers(self):
        self.register_use_case_handler(
            "https://gematik.de/fhir/atf/CodeSystem/operation-identifier-cs",
            "atf;Empfangsbestaetigung",
            self.empfangsbestaetigungs_handler
        )

        self.register_use_case_handler(
            "https://gematik.de/fhir/atf/CodeSystem/service-identifier-cs",
            "Selbsttest;Lieferung",
            SelbsttestLieferungValidator(self.sender, self.source)
        )

    def register_use_case_handler(self, system: str, code: str, handler: BaseUseCaseValidator):
        self.use_case_handlers[(system, code)] = handler

    def process_bundle(self, bundle: Bundle) -> BundleContent:
        parsed_bundle = self.parse_bundle(bundle)

        if parsed_bundle is None:
            return

        if not self.is_valid_atf_message(parsed_bundle):
            return

        message_header = self.get_message_header(parsed_bundle)

        if message_header is None:
            logger.error(
                "Die empfangene Nachricht ist keine gültige ATF-Nachricht. Ein MessageHeader fehlt.")
            return
        service_identifier = ServiceIdentifier(message_header.eventCoding.system,
                                               message_header.eventCoding.code,
                                               message_header.eventCoding.display)

        if (service_identifier.system, service_identifier.code) in self.use_case_handlers:
            handler = self.use_case_handlers[(service_identifier.system, service_identifier.code)]
            ressources, issues = handler.handle(message_header, bundle)

            if issues:
                self.send_empfangsbestätigung(message_header, issues)

            if ressources:
                focusRessource = BundleContent(
                    service_identifier, message_header.focus, message_header.sender.identifier.value, ressources)
                self.bundle_content_to_process_event.trigger(focusRessource)

        else:
            issues = self.create_unsupported_use_case_issues(
                message_header.eventCoding.code)
            self.send_empfangsbestätigung(message_header, issues)
            return

    def parse_bundle(self, bundle: Bundle):
        try:
            return Bundle.parse_raw(bundle.json())
        except JSONDecodeError:
            logger.error(
                "Die empfangene Nachricht ist keine gültige FHIR-Nachricht.")
            return None

    def is_valid_atf_message(self, parsed_bundle: Bundle):
        bundle_codesystem = parsed_bundle.meta.profile[0]
        if bundle_codesystem != "https://gematik.de/fhir/atf/StructureDefinition/bundle-app-transport-framework":
            logger.error(
                "Die empfangene Nachricht ist keine gültige ATF-Nachricht.")
            return False

        return True

    def get_message_header(self, parsed_bundle: Bundle):
        return next((entry.resource for entry in parsed_bundle.entry if isinstance(
            entry.resource, MessageHeader)), None)

    def create_unsupported_use_case_issues(self, event_code: str):
        return [
            OperationOutcomeIssue(
                severity="error",
                code="processing",
                diagnostics=f"Die empfangene Nachricht mit dem {event_code} kann nicht verarbeitet werden, da der Use-Case nicht unterstützt wird."
            )
        ]

    def send_empfangsbestätigung(self, message_header, issues):
        operation_outcome = OperationOutcomeCreator.create_operation_outcome_ressource(
            message_id=message_header.id,
            issues=issues)

        message_to_send = self.create_MessageToSend(
            message_header, operation_outcome)

        self.message_to_send_event.trigger(message_to_send)

    def create_MessageToSend(self, message_header, operation_outcome):
        destination = [MessageHeaderDestination(endpoint=message_header.source.endpoint,
                                                receiver=message_header.sender)]

        operationOutcomeBundle = OperationOutcomeBundleCreator.create_operation_outcome_receipt_bundle(
            self.sender, self.source, destination, operation_outcome)
        message_to_send = MessageToSend(
            atf_bundle=operationOutcomeBundle,
            receiver=message_header.sender.identifier.value,
            message_type="atf;Empfangsbestaetigung"
        )

        return message_to_send



from uuid import uuid4
from fhir.resources.messageheader import MessageHeaderDestination, MessageHeaderSource
from fhir.resources.bundle import BundleEntry
from fhir.resources.communication import Communication, CommunicationPayload
from fhir.resources.fhirtypes import ReferenceType, CodeableConceptType
from fhir.resources.attachment import Attachment
from app_transport_framework_library.ressource_creators.message_bundle_creator import MessageBundleCreator

from app_transport_framework_library.ressource_creators.message_header_creator import MessageHeaderCreator



class CommunicationCreator:
    def __init__(self):
        pass

    def create(id: str, status: str, priority: str, subject_display: str, sender_display: str,
               content_attachment_title: str, content_attachment_content_type: str,
               content_attachment_data: str, sent: str, received: str) -> Communication:
        communication = Communication(
            id=id,
            status=status,
            priority=priority,
            subject=ReferenceType(display=subject_display),
            sender=ReferenceType(display=sender_display),
            payload=[
                CommunicationPayload(
                    contentAttachment=Attachment(
                        title=content_attachment_title,
                        contentType=content_attachment_content_type,
                        data=content_attachment_data
                    )
                )
            ],
            sent=sent,
            received=received
        )
        category_coding = CodeableConceptType(
            system="http://loinc.org",
            code="45012-6",
            display="Communication regarding test results"
        )
        communication.category = [
            CodeableConceptType(coding=[category_coding])]
        return communication


class TestMessageCreator:
    def create_test_bundle(sender: ReferenceType, source: MessageHeaderSource,   receiver: ReferenceType, message_id: str):
        communication_id = str(uuid4())

        source = source
        message_receiver = receiver
        destination = [MessageHeaderDestination(endpoint="https://receiver.example.com/endpoint",
                                                receiver=message_receiver)]

        message_header = MessageHeaderCreator.create_message_header(
            message_id,
            sender,
            source,
            destination,
            code_system="https://gematik.de/fhir/atf/CodeSystem/service-identifier-cs",
            use_case="Selbsttest;Lieferung",
            use_case_display="Diese Dienstkennung dient ausschließlich der Einrichtung des Kontos innerhalb eines PVS und des Testes, ob Nachrichten versendet und empfangen werden können. Diese Dienstkennung wird im PVS bei der normalen Abholung von Nachrichten ignoriert.",
            focus_reference=communication_id)

        communication = CommunicationCreator.create(
            id=communication_id,
            status="completed",
            priority="routine",
            subject_display="Max Mustermann",
            sender_display="Dr. Anna Schmidt",
            content_attachment_title="Selbsttest Bestätigung",
            content_attachment_content_type="text/plain",
            content_attachment_data="U2VsYnN0dGVzdCBhYnNjaGxpZXNzZW4uIEJpdHRlIGtsYXJlbiBTaWUgZGllIFRlc3RlcmdlYm5pc3NlIGFiLg==",
            sent="2023-03-29T13:28:17.239+02:00",
            received="2023-03-29T13:30:00.000+02:00"
        )

        resources = [
            BundleEntry(
                fullUrl=f"urn:uuid:{communication_id}",
                resource=communication
            )
        ]

        bundle = MessageBundleCreator.create_message_bundle(
            message_header=message_header,
            resources=resources
        )

        return bundle



from fhir.resources.bundle import Bundle
from fhir.resources.operationoutcome import OperationOutcomeIssue


class Empfangsbestaetigung:
    def __init__(self, message_id: str, message_received: bool, sender: str, receiver: str, details: OperationOutcomeIssue):
        self.message_id = message_id
        self.message_received = message_received
        self.receiver = receiver
        self.sender = sender
        self.details = details



from fhir.resources.meta import Meta
from fhir.resources.messageheader import MessageHeader, MessageHeaderDestination, MessageHeaderSource
from fhir.resources.fhirtypes import ReferenceType
from fhir.resources.coding import Coding
from typing import List


class MessageHeaderCreator:
    @staticmethod
    def create_message_header(
        id: str,
        message_sender: ReferenceType,
        source: MessageHeaderSource,
        destinations: List[MessageHeaderDestination],
        code_system: str,
        use_case: str,
        use_case_display: str,
        focus_reference: str,
    ) -> MessageHeader:
        message_header = MessageHeader(
            id=id,
            meta=Meta.construct(profile=[
                "https://gematik.de/fhir/atf/StructureDefinition/message-header-app-transport"
            ]),
            eventCoding=Coding(
                system=code_system,
                code=use_case,
                display=use_case_display
            ),
            source=source,
            destination=destinations,
            sender=message_sender,
            focus=[ReferenceType(reference=f"urn:uuid:{focus_reference}")]
        )
        return message_header
